(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(*  This example is based on an example designed for OCamlP3l.         *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rdec2.mli,v 1.9 2017/01/04 11:58:30 fclement Exp $ *)

(** {3 2D regular rectangular grids}

 Module for the representation of 2D regular rectangular grids with regular
 domain decomposition capabilities.
 Overlapping of the domain decomposition is parameterized.

 A regular_decomposition record contains:
   a computational_rectangular_grid record:
     a rectangular_grid record:
       a rectangle record:
         x0 and y0, coordinates of the lower-left corner of the rectangle
         lx and ly, length of the rectangle in both directions
       nx and ny, number of intervals in both directions
     px and py, half the overlapping of the subdomains in both directions
   nsbdx and nsbdy, number of subdomains in both directions.

 A valid regular_decomposition must satisfy:
   lx > 0. and ly > 0.
   nx > 0 and ny > 0
   px > 0 and py > 0
   px <= nx / nsbdx and py <= ny / nsbdy
   nsbdx > 0 and nsbdy > 0
   nsbdx <= nx and nsbdy <= ny
   nx mod nsbdx = 0 and ny mod nsbdy = 0. *)

(** {3 Regular decomposition} *)

type length = Upervasives.length
and index = Upervasives.index
(** Aliases of type Unatural.t. *)
;;

type size = length * length;;

type rectangle =
  {
   x0 : float; y0 : float;
   lx : float; ly : float;
  }

and rectangular_grid =
  {
   domain : rectangle;
   nx : length; ny : length;
  }

and computational_rectangular_grid =
  {
   mesh : rectangular_grid;
   px : length; py : length;
  }

and regular_decomposition =
  {
   comp_grid : computational_rectangular_grid;
   nsbdx : length; nsbdy : length;
  }

and t = regular_decomposition
;;

val number_of_subdomains_of_regular_decomposition : t -> size;;

val print_computational_rectangular_grid :
  out_channel -> computational_rectangular_grid -> unit
;;

exception Wrong_Input of string;;

val parse_rdec_config_file : string -> t;;

val check_regular_decomposition : t -> unit;;

(** {3 Subdomain} *)

type subdomain_index = index

and subdomain =
  {
   dec : t;
   number : subdomain_index;
  }
;;

exception One_d_index_out_of_range of index;;
exception Two_d_index_out_of_range of index * index;;

val two_indices_numbering : length -> length -> index -> index * index;;
val one_index_numbering : length -> length -> index * index -> index;;

val rectangular_grid_of_subdomain : subdomain -> rectangular_grid;;

val neighbors_of : subdomain -> subdomain_index list;;

val computational_rectangular_grid_of_subdomain :
    subdomain -> computational_rectangular_grid
;;

(** {3 Connectivity table} *)

type neighbor = index

and neighborhood =
  {
   east_n : neighbor option;
   north_n : neighbor option;
   south_n : neighbor option;
   west_n : neighbor option;
  }

and connectivity_table = neighborhood array
;;

val connectivity_table_of_regular_decomposition : t -> connectivity_table;;

val east_neighbor : connectivity_table -> index -> neighbor option;;
val north_neighbor : connectivity_table -> index -> neighbor option;;
val south_neighbor : connectivity_table -> index -> neighbor option;;
val west_neighbor : connectivity_table -> index -> neighbor option;;

val print_connectivity_table : out_channel -> connectivity_table -> unit;;

(** {3 Interface} *)

type edge_values = float array

and boundary =
  {
   east_ev : edge_values;
   north_ev : edge_values;
   south_ev : edge_values;
   west_ev : edge_values;
  }

and boundary_size =
  {
   size_x : Unatural.t;
   size_y : Unatural.t;
  }

and interface = boundary array
;;

val proj_boundary : subdomain -> boundary -> (index * boundary) -> boundary;;
val normalize_boundary : boundary -> boundary;;
val print_boundary : out_channel -> boundary -> unit;;
val read_boundary : Scanf.Scanning.scanbuf -> boundary_size -> boundary;;
val size_of_boundary : boundary -> boundary_size;;

val make_zero_boundary : boundary_size -> boundary;;
val make_zero_interface : t -> interface;;
val projection : connectivity_table -> interface -> interface;;

val for_all_subdomains : length * length -> (index -> unit) -> unit;;
