(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.23 2017/05/15 13:58:10 fclement Exp $ *)

module Nat = Unatural;;
module A = Uarray;;
module Co = Ucommand_option;;

let _0_ = Nat._0_;;

open Skl;;
open Options;;

let stream_of_matrix mat =
  let count = ref _0_
  and mx = A.length mat in
  let gen () =
    if !count >= mx then None else
    (* Now, we have !count < mx = A.length mat.
       Therefore, the next A.unsafe_get will be safe. *)
    let line = A.unsafe_get mat !count in
    Nat.incr count;
    Some line in
  Sklstream.of_fun gen
;;

let vector_of_stream s = A.of_list (Sklstream.to_list s);;

let predicate niter =
  let count = ref niter in
  fun _ -> decr count; !count <> 0
;;

let display x =
  Helper.IO.output_raw stdout x;
  Printf.printf "\n%!";
  x
;;

(* The atomic skeletons. *)

(*let scalar_product_skl = skl v0 -> Helper.scalar_prod v0;;*)
let scalar_product_skl = Skl.make_skl Helper.scalar_prod;;

(*let mk_pvect_skl = skl () -> Tridiag.mk_pvect;;*)
let mk_pvect_skl = Skl.make_skl (fun () -> Tridiag.mk_pvect);;

(*let pmult_skl = skl tmat -> Tridiag.pmult tmat;;*)
let pmult_skl = Skl.make_skl Tridiag.pmult;;

(*let predicate_skl = skl niter -> predicate niter;;*)
let predicate_skl = Skl.make_skl predicate;;

(*let display_skl = skl () -> display;;*)
let display_skl = Skl.make_skl (fun () -> display);;

(* The full computational skeleton expressions. *)

let matvec_skl vec nw = farm (scalar_product_skl vec, nw);;

let tridiag_skl nw niter tmat =
  loop
    (predicate_skl niter, (* do { } while (--count <> 0); *)
     mk_pvect_skl () |||
     farm_vector (pmult_skl tmat, nw) |||
     display_skl ())
;;

let main { Parfun.parfun = parfun; } =
  Co.parse ();
  let nw = Sklml_options.get_number_of_workers () in
  let { Options.vector_file = vf;
        matrix_file = mf;
        running_mode = mode; } = Options.get_options () in
  match mode with
  (* First demo, simple parallel multiplication *)
  | Std ->
    let (mat, vec) = Helper.fetch_std_data mf vf in
    let input_stream = stream_of_matrix mat in
    let output_stream = parfun (matvec_skl vec nw) input_stream in
    Helper.IO.output_vector stdout (vector_of_stream output_stream);
  (* Second demo, triangular matrix multiplication iterated *)
  | Tri ->
    let (mat, vec) = Helper.fetch_tri_data mf vf in
    let is = Sklstream.singleton (Tridiag.split_vect 5 vec) in
    let os = parfun (tridiag_skl nw 15 mat) is in
    Sklstream.iter ignore  os;
;;

Parfun.pardo main;;
