(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.37 2017/05/15 13:58:10 fclement Exp $ *)

module Nat = Unatural;;
module B = Ubytes;;
module Buf = Ubuffer;;
module Co = Ucommand_option;;

let _1024_ = Nat.unsafe_of_int 1024;;

let rec read_lines ic ob =
  let buf = B.create _1024_ in
  let len = input ic buf 0 1024 in
  if len = 0 then raise End_of_file else
  Buf.add_subbytes ob buf 0 len;
  read_lines ic ob
;;

let eval_command cmd args =
  let cmd_line = cmd ^ " " ^ args in
  let ic = Unix.open_process_in cmd_line in
  let ob = Buffer.create 1024 in
  try read_lines ic ob with
  | End_of_file ->
    let status = Unix.close_process_in ic in
    let output = Buffer.contents ob in
    Buffer.clear ob;
    (args, status, output)
;;

(* The atomic eval command skeleton. *)
(*let eval_command_skl = skl cmd -> eval_command cmd;;*)
let eval_command_skl = Skl.make_skl eval_command;;

(* The full computational skeleton expression. *)
let parexec_skl nw cmd = Skl.farm (eval_command_skl cmd, nw);;

let open_file = function
  | "-" -> stdin
  | fname -> open_in fname
;;

let print_result cmd (args, status, output) =
  Printf.printf "%s" output;
  match status with
  | Unix.WEXITED n ->
    if n <> 0 then
      Printf.eprintf
        "Command '%s' with arguments '%s' failed with status %d.\n"
        cmd args n
  | Unix.WSIGNALED n
  | Unix.WSTOPPED n ->
    Printf.eprintf
      "Command '%s' with arguments '%s' was killed/stopped by signal %d.\n"
      cmd args n
;;

let main { Parfun.parfun = parfun; } =
  Co.parse ();
  let command = Options.get_command () in
  let ic = open_file (Options.get_command_arguments ()) in
  let nw = Sklml_options.get_number_of_workers () in
  let is = Sklstream.of_channel_by_lines ic in
  let os = parfun (parexec_skl nw command) is in
  Sklstream.iter (print_result command) os
;;

Parfun.pardo main;;
