(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: worker.ml,v 1.9 2017/01/04 11:58:31 fclement Exp $ *)

module Nat = Unatural;;
module Perv = Upervasives;;
module L1 = Ulist1;;
module A = Uarray;;
module D = Udebug;;

let _0_ = Nat._0_;;

open Misc_sprint;;

let qual = Perv.qualify "Worker_lib";;
let failwith = Perv.failwith_qual qual;;

(* Tasks. *)

type init_engine = Pio.spf -> unit -> int list
and comp_engine = Pio.spf -> int -> int
and final_engine = Pio.spf -> unit -> unit
;;

type task_engines =
  {
   init_engine : init_engine;
   light_comp_engine : comp_engine;
   heavy_comp_engine : comp_engine;
   final_engine : final_engine;
  }
;;

type tasks =
  {
   init : (unit, int list) Pio.task;
   light_comp : (int, int) Pio.task;
   heavy_comp : (int, int) Pio.task;
   final : (unit, unit) Pio.task;
  }
;;

let make_init engine ({ D.spf_debug = debug; _; } as spf) () =
  let l = engine spf () in
  debug (sprint_int_list "l" l);
  l
;;

let make_comp engine ({ D.spf_debug = debug; _; } as spf) i =
  let i' = engine spf i in
  debug (sprint_int "i'" i');
  i'
;;

let make_final engine = engine;;

let make_tasks task_engines =
  let init = make_init task_engines.init_engine in
  let light_comp = make_comp task_engines.light_comp_engine in
  let heavy_comp = make_comp task_engines.heavy_comp_engine in
  let final = make_final task_engines.final_engine in
  {
   init = init;
   light_comp = light_comp;
   heavy_comp = heavy_comp;
   final = final;
  }
;;

(* Wrappers. *)

let mk_init_proj { D.spf_report = report; _; } vals =
  let nargs = A.length vals in
  if nargs <> Nat._0_ then
    failwith "init_proj" "initialization should receive zero value" else
  report "zero value received";
;;

let mk_init_inj _spf l = [| Typed_value.of_int_vector (A.of_list l); |];;

let make_init_wrapper =
  Pio.make_task_wrapper "init" mk_init_inj mk_init_proj
;;

let mk_comp_proj { D.spf_report = report; _; } vals =
  let nargs = A.length vals in
  if nargs <> Nat._1_ then
    failwith "comp_proj" "compute should receive one value" else
  (* Now, we have nargs = A.length vals = 1.
     Therefore, the next A.unsafe_get will be safe. *)
  report "one value received";
  let i = Typed_value.to_int (A.unsafe_get vals _0_) in
  report "integer extracted";
  i
;;

let mk_comp_inj _spf i = [| Typed_value.of_int i; |];;

let make_light_comp_wrapper =
  Pio.make_task_wrapper "light comp" mk_comp_inj mk_comp_proj
;;

let make_heavy_comp_wrapper =
  Pio.make_task_wrapper "heavy comp" mk_comp_inj mk_comp_proj
;;

let mk_final_proj { D.spf_report = report; _; } vals =
  let nargs = A.length vals in
  if nargs <> Nat._0_ then
    failwith "final_proj" "finalization should receive zero value" else
  report "zero value received";
;;

let mk_final_inj _spf () = [| Typed_value.of_string "OK, bye!"; |];;

let make_final_wrapper =
  Pio.make_task_wrapper "final" mk_final_inj mk_final_proj
;;

(* Task computation functions. *)

let make_initialize t pf piob =
  Pio.worker_wrap_task piob (make_init_wrapper pf "init") t
;;

let make_compute_light t pf piob =
  Pio.worker_wrap_task piob (make_light_comp_wrapper pf "light_comp") t
;;

let make_compute_heavy t pf piob =
  Pio.worker_wrap_task piob (make_heavy_comp_wrapper pf "heavy_comp") t
;;

let make_finalize t pf piob vals =
  Pio.worker_wrap_task piob (make_final_wrapper pf "final") t vals;
  (raise Exit : unit);
;;

(* Main loop. *)

let welcome_action () = "start user model"
and farewell_action () = "finish user model"
;;

let mk_task_tbl tasks =
  L1.of_list
    [
     ("init", make_initialize tasks.init);
     ("light comp", make_compute_light tasks.light_comp);
     ("heavy comp", make_compute_heavy tasks.heavy_comp);
     ("final", make_finalize tasks.final);
    ]
;;

let main_loop tasks =
  let task_tbl = mk_task_tbl tasks in
  Pio.worker_main welcome_action farewell_action task_tbl
;;
