(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: make_domain.ml,v 1.24 2017/01/04 11:58:28 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;
module A = Uarray;;

let _0_ = Nat._0_;;

open Tools_common_types;;
open Skl;;

type 'a border = worker_index * 'a
and 'a borders = 'a border list
;;

type worker_connectivity = worker_index list
and connectivity_table = worker_connectivity array
;;

type ('a, 'b) worker_spec =
  ('a borders, 'a * 'b) Skl.t * worker_connectivity
;;

let from_some = function
  | None -> invalid_arg "from_some"
  | Some x -> x
;;

let array_split a =
  let l = A.length a in
  if l = _0_ then ([||], [||]) else
  (* In the following, we will have 0 <= i < l = A.length a.
     Therefore, the next A.unsafe_get will be safe. *)
  let fx i = fst (A.unsafe_get a i)
  and fy i = snd (A.unsafe_get a i) in
  (A.init l fx, A.init l fy)
;;

let sanity_check speca =
  let n = A.length speca in
  let check_conn = L.for_all (( > ) n) in
  let anding f x y = x && f y in
  A.fold_left (anding (fun (_, l) -> check_conn l)) true speca
;;

(* connectivity_table -> (('a * 'b) array, 'a borders array) Skl.t *)
let mux =
  let f_mux conn_spec a =
    (* The connectivity spec conn_spec is assumed to connect indices from
       array a.
       Therefore, the next A.unsafe_get will be safe. *)
    A.map (L.map (fun i -> (i, fst (A.unsafe_get a i)))) conn_spec in
  let mux_pure conn_spec = Sklstream.map (f_mux conn_spec) in
  ignore (Expor.register mux_pure);
  fun conn_spec -> atom (conn_spec, from_some (Expor.is_registered mux_pure))
;;

(* unit -> (('a * 'b) array, ('a * 'b option) array) Skl.t *)
let demux =
  let f_demux (x, y) = (x, Some y) in
  let demux_pure () = Sklstream.map (A.map f_demux) in
  ignore (Expor.register demux_pure);
  fun () -> atom ((), from_some (Expor.is_registered demux_pure))
;;

(* unit -> (('a * 'b option) array, 'b array) Skl.t *)
let ppcond =
  let f_ppcond (_, y) = from_some y in
  let ppcond_pure () = Sklstream.map (A.map f_ppcond) in
  ignore (Expor.register ppcond_pure);
  fun () -> atom ((), from_some (Expor.is_registered ppcond_pure))
;;

(* unit -> ('a array, ('a * 'b option) array) Skl.t *)
let pre =
  let f_pre x = (x, None) in
  let pre_pure () = Sklstream.map (A.map f_pre) in
  ignore (Expor.register pre_pure);
  fun () -> atom ((), from_some (Expor.is_registered pre_pure))
;;

(* unit -> (('a * 'b option) array, ('a * 'b) array) Skl.t *)
let post =
  let f_post (x, y) = (x, from_some y) in
  let post_pure () = Sklstream.map (A.map f_post) in
  ignore (Expor.register post_pure);
  fun () -> atom ((), from_some (Expor.is_registered post_pure))
;;

let gen_mk_domain speca pskl =
  if not (sanity_check speca) then
    invalid_arg "mk_domain: wrong input specification" else
  let (skls, conn_spec) = array_split speca in
  pre () |||
  loop (pskl, mux conn_spec ||| rails skls ||| demux ()) |||
  post ()
;;

let mk_domain speca pskl =
  let pskl = ppcond () ||| pskl in
  gen_mk_domain speca pskl
;;

let mk_f_domain speca p =
  let pskl_reg =
    Expor.register (fun () ->
      Sklstream.map (fun a -> p (A.map (fun (_, y) -> from_some y) a))) in
  let pskl = atom ((), pskl_reg) in
  gen_mk_domain speca pskl
;;

(* unit -> ('a array, ('a * 'b) array) Skl.t *)
let pre_opt =
  let f_pre_opt a =
    let magic_default = Obj.magic 0 in
    A.map (fun x -> (x, magic_default)) a in
  let pre_opt_pure () = Sklstream.map f_pre_opt in
  ignore (Expor.register pre_opt_pure);
  fun () -> atom ((), from_some (Expor.is_registered pre_opt_pure))
;;

(* unit -> (('a * 'b) array, 'b array) Skl.t *)
let ppcond_opt =
  let f_ppcond_opt = snd in
  let ppcond_opt_pure () = Sklstream.map (A.map f_ppcond_opt) in
  ignore (Expor.register ppcond_opt_pure);
  fun () -> atom ((), from_some (Expor.is_registered ppcond_opt_pure))
;;

let gen_mk_domain_opt speca pskl =
  if not (sanity_check speca) then
    invalid_arg "mk_domain: wrong input specification" else
  let (skls, conn_spec) = array_split speca in
  pre_opt () |||
  loop (pskl, mux conn_spec ||| rails skls)
;;

let mk_domain_opt speca pskl =
  gen_mk_domain_opt speca (ppcond_opt () ||| pskl)
;;

let mk_f_domain_opt speca p =
  let pskl_reg =
    Expor.register (fun () ->
      Sklstream.map (fun a -> p (A.map snd a))) in
  let pskl = atom ((), pskl_reg) in
  gen_mk_domain_opt speca pskl
;;

(* TODO: Maybe only add a default *)
