(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: master.ml,v 1.54 2017/01/04 11:58:29 fclement Exp $ *)

module Nat = Unatural;;
module L = Ulist;;

let _0_ = Nat._0_;;

open Tools_common_types;;
open Daemon_types;;
open Node_types;;

let { tagged_debug = debug; _; } = Options.services;;

let debug fname = debug (Printf.sprintf "Master.%s" fname);;

type slave =
  {
   io : IO.t;
   addr : Unix.inet_addr;
   col : color;
  }
;;

type waiting_state =
  {
   slaves_seen : slave list;
   slave_count : number_of_workers;
   listen_sock : Unix.file_descr;
  }
;;

type serving_state =
  {
   slaves : slave list;
   sklml_io : IO.t;
  }
;;

type state =
  | Start
  | Waiting_slaves of waiting_state
  | Start_serving of slave list
  | Serving of serving_state
  | Stop
;;

type request =
  | Slave_port_available of Unix.file_descr
  | Slave_connect of slave * waiting_state
  | Client_available of serving_state
  | Client_request of Skl_comm.request * serving_state
  | Client_lost
;;

type placement_policy =
  | First_fit (* Choose the first fitting bin *)
  | Swap (* Used when the job can't be assigned (colors become negative) *)
  | Worst_fit (* Choose the emptiest bin *)
;;
(* Placement policies used to assign a job to a processor, here, they
  are sorted by speed. *)

(* val update : ('a -> 'a) -> 'a -> 'a list -> 'a list;;
 [update f a l] updates field [a] of [l] using function [f].
 Like [map], on a single element. *)
let update f a l =
  let rec loop acc = function
    | [] -> invalid_arg "Master.loop"
    | h :: t ->
      if h == a then L.rev_append acc (f h :: t) else
      loop (h :: acc) t in
  loop [] l
;;

(* val extract : 'a -> 'a list -> 'a list;;
 [extract a l] updates list [l] by removing element [a].
 If it is not in the list it fails with an [Invalid_arg] exception. *)
let extract a l =
  let rec loop acc l =
    match l with
    | [] -> invalid_arg "Master.loop"
    | h :: t ->
      if h == a then L.rev_append acc t else
      loop (h :: acc) t in
  loop [] l
;;

let rec in_max cmp m = function
  | [] -> m
  | h :: t -> in_max cmp (if cmp h m then h else m) t
;;

(* val max : ('a -> 'a -> bool) -> 'a list -> 'a;;
 [max cmp l] returns the biggest element of list [l] using order [cmp].
 [cmp a b] is true iff '[a] > [b]'. *)
let max cmp = function
  | [] -> invalid_arg "Master.max"
  | h :: t -> in_max cmp h t
;;

(* val placement : placement_policy -> slave list -> color -> slave Ulist1.t *)
let rec placement policy slaves job =
  let update_slave s =
    (* Beware: col may become negative... *)
    let new_col = s.col - job in
    debug "placement"
      (Printf.sprintf "Master daemon: s.col = %d, job = %d, new_col = %d"
        s.col job new_col);
    { s with col = new_col; } in
  match policy with
  | First_fit ->
    let rec loop acc = function
      | [] -> placement Swap slaves job (* we failed *)
      | ({ col = c; _ } as slave) :: t ->
        if c < job then loop (slave :: acc) t else
        (slave, L.rev_append acc (update_slave slave :: t)) in
    loop [] slaves
  | Worst_fit
  | Swap ->
    let compare a b = a.col > b.col in
    let slave = max compare slaves in
    let slaves = update update_slave slave slaves in
    (slave, slaves)
;;

(* val req_generator : master_node -> state -> request option *)
let req_generator node s =
  try
    match s with
    | Start ->
      let (sock, addr) = IO.create_bound_socket (Some node.mn_slave_port) in
      Log.say Log.Say "listenning for slaves on %a" IO.dump_sockaddr addr;
      Some (Slave_port_available sock)
    | Waiting_slaves state ->
      let ((ic, _ as io), addr) =
        IO.wait_connection state.listen_sock false in
      let addr =
        if addr = Unix.inet_addr_loopback then node.mn_local_address else
        addr in
      Log.say Log.Say "slave accepted (%s)" (Unix.string_of_inet_addr addr);
      let { color = col; } = IO.of_channel ic in
      Some (Slave_connect ({ col; io; addr; }, state))
    | Start_serving slaves ->
      let (sock, addr) = IO.create_bound_socket (Some node.mn_sklml_port) in
      Log.say
        Log.Say "listenning for sklml program on %a" IO.dump_sockaddr addr;
      let (io, _) = IO.wait_connection sock true in
      Some (Client_available { sklml_io = io; slaves = slaves; })
    | Serving ({ sklml_io = (ic, _); _ } as state) ->
      let request =
        match
          try Some (Skl_comm.extract_request ic) with
          | End_of_file -> None
        with
        | None -> Client_lost
        | Some r -> Client_request (r, state) in
      Some request
    | Stop -> None
  with
  | IO.IOError msg ->
    Log.say Log.Err "IO error: %s" msg;
    None
;;

(* val req_handler : master_node -> request -> state *)
let req_handler node = function
  | Slave_port_available fd ->
    let state =
      {
       slaves_seen = [];
       slave_count = _0_;
       listen_sock = fd;
      } in
    Waiting_slaves state
  | Slave_connect (slave, waiting_state) ->
    if Nat.succ waiting_state.slave_count = node.mn_slave_number then begin
      Unix.close waiting_state.listen_sock;
      Start_serving (slave :: waiting_state.slaves_seen)
    end else
      Waiting_slaves
        { waiting_state with
          slaves_seen = slave :: waiting_state.slaves_seen;
          slave_count = Nat.succ waiting_state.slave_count; }
  | Client_available serving_state -> Serving serving_state
  | Client_request (req, serving_state) ->
    let { slaves = slaves; sklml_io = (_, oc); } = serving_state in
    (* maybe_rep is
      Some (daemon_rep, slave_addr, new_slaves, slaves) if the
      request was handled by a slave, and
      None if no more slaves are available *)
    (* slaves must be returned in case of a slave loss followed
      by a Spawn_failure reply. new_slaves is slaves updated (the
      color of the chosen slave is changed). slave_addr is the
      address of the chosen slave. daemon_rep is the reply of the
      slave. *)
    let maybe_rep =
      let rec send_request slaves =
        if slaves = [] then None else
        let (daemon_req, slave, new_slaves) =
          match req with
          | Skl_comm.Atom_req (p, atom_id, col) ->
            let (slave, slaves) = placement First_fit slaves col in
            (Spawn_atom (p, atom_id), slave, slaves)
          | Skl_comm.Loop_req (t1, t2) ->
            let (slave, _) = placement Worst_fit slaves 0 in
            (Spawn_loop (t1, t2), slave, slaves)
          | Skl_comm.Pipe_req (t1, t2) ->
            let (slave, _) = placement Worst_fit slaves 0 in
            (Spawn_pipe (t1, t2), slave, slaves) in
        match
          try Some (Spawnlib.request_slave slave.io daemon_req) with
          | End_of_file
          | IO.IOError _ -> None
        with
        | Some rep -> Some (rep, slave.addr, new_slaves, slaves)
        | None ->
          Log.say Log.Warn
            "slave lost, continuing without him (ip:%s)"
            (Unix.string_of_inet_addr slave.addr);
          send_request (extract slave slaves) in
      send_request slaves in
    Spawnlib.maybe maybe_rep
      (* error *)
      (fun () ->
        Log.say Log.Err "no more slaves remaining, exiting";
        Stop)
      (* success *)
      (fun (daemon_rep, slave_addr, new_slaves, slaves) ->
        let (skl_reply, slaves) =
          match daemon_rep with
          | Spawn_success { worker_port = p; } ->
            (Skl_comm.Success (Unix.ADDR_INET (slave_addr, Nat.to_int p)),
             new_slaves)
          | Spawn_failure -> (Skl_comm.Failure, slaves) in
        Skl_comm.send_reply oc skl_reply;
        Serving { serving_state with slaves = slaves; })
  | Client_lost -> Stop
;;

(* val master_loop : master_node -> state -> state *)
let master_loop node s =
  Spawnlib.master_loop (req_generator node) (req_handler node) s
;;

let fail_msg = Printf.sprintf "System failure in %S state.";;

let start node =
  match master_loop node Start with
  | Stop -> ()
  | Start -> Spawnlib.fail (fail_msg "Start")
  | Waiting_slaves _ -> Spawnlib.fail (fail_msg "Waiting_slaves")
  | Start_serving _ -> Spawnlib.fail (fail_msg "Start_serving")
  | Serving _ -> Spawnlib.fail (fail_msg "Serving")
;;
