(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: slave.ml,v 1.56 2017/01/04 11:58:29 fclement Exp $ *)

module Nat = Unatural;;
module S = Ustring;;
module A = Uarray;;
module U = Uunix;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
;;

open Tools_common_types;;
open Daemon_types;;
open Node_types;;

let { tagged_debug = debug; _; } = Options.services;;

let debug fname = debug (Printf.sprintf "Slave.%s" fname);;

(* val to_tree : ('a, 'b) Iotree.ioatree -> 'a Sklstream.t -> 'b Sklstream.t *)
let to_tree tree is =
  let tree = Iotree.iotree_of_ioatree tree in
  match IO.fork () with
  | IO.Child -> Iotree.send_in_tree tree is; exit 0
  | IO.Parent _ -> Iotree.receive_in_tree tree
;;

(* val spawn_atom : string -> Spawnlib.closure_id -> Spawnlib.spawn_return *)
let spawn_atom program_path (params, closure_id) =
  let debug = debug "spawn_atom" in
  let (cmd, args) =
    (program_path, [| program_path; "-atom"; Nat.to_string closure_id; |]) in
  debug
    (Printf.sprintf "command line = %S"
       (S.concat " " (A.to_list args)));
  try
    match IO.fork () with
    | IO.Child -> U.execvp cmd args
    | IO.Parent pid ->
      let (fifo_in, fifo_out) = Skl_comm.open_skl_fifo true pid in
      IO.to_channel fifo_out params true;
      let port = IO.of_channel fifo_in in
      Skl_comm.clean_skl_fifo pid;
      Log.say (Log.Log _2_) "Spawn atom success %d" port;
      Spawn_success { worker_port = port; }
  with
  | IO.IOError msg ->
    Log.say Log.Err "IO error is Slave.spawn_atom: %s" msg;
    Spawn_failure
  | End_of_file
  | Scanf.Scan_failure _ ->
    Log.say Log.Err "Scanf.bscanf error in Slave.spawn_atom";
    Spawn_failure
;;

(* val spawn_loop :
  ('a, bool) Iotree.ioatree ->
  ('a, 'a) Iotree.ioatree -> Spawnlib.spawn_return *)
let spawn_loop iotreep iotree =
  let spawnf (ic, oc) (ioatreep, ioatree) =
    let treep = Iotree.iotree_of_ioatree ioatreep in
    let tree = Iotree.iotree_of_ioatree ioatree in
    let p = Iotree.map_tree treep in
    let f = Iotree.map_tree tree in
    Sklstream.to_io_channel
      (Sklstream.loop p f (Sklstream.of_io_channel ic)) oc in
  try
    let port = Spawnlib.spawn_sock spawnf (iotreep, iotree) in
    Log.say (Log.Log _2_) "Spawn loop success %d" (Nat.to_int port);
    Spawn_success { worker_port = port; }
  with
  | IO.IOError msg ->
    Log.say Log.Err "IO error in Slave.spawn_loop: %s" msg;
    Spawn_failure
;;

(* val spawn_pipe :
  ('a, 'b) Iotree.ioatree ->
  ('b, 'c) Iotree.ioatree -> Spawnlib.spawn_return *)
let spawn_pipe iotree1 iotree2 =
  let spawnf (ic, oc) (iotree1, iotree2) =
    let os1 = to_tree iotree1 (Sklstream.of_io_channel ic) in
    let os2 = to_tree iotree2 os1 in
    Sklstream.to_io_channel os2 oc in
  try
    let port = Spawnlib.spawn_sock spawnf (iotree1, iotree2) in
    Log.say (Log.Log _2_) "Spawn pipe success %d" (Nat.to_int port);
    Spawn_success { worker_port = port; }
  with
  | IO.IOError msg ->
    Log.say Log.Err "IO error in Slave.spawn_pipe: %s" msg;
    Spawn_failure
;;

(* val slave_loop : ic * oc -> node -> unit *)
let slave_loop io { sn_sklml_program_path = path; _; } =
  let req_handler = function
    | Spawn_atom (params, id) -> spawn_atom path (params, id)
    | Spawn_loop (tree1, tree2) -> spawn_loop tree1 tree2
    | Spawn_pipe (tree1, tree2) -> spawn_pipe tree1 tree2 in
  Spawnlib.slave_loop io req_handler
;;

let rec chomp s =
  let len = S.length s in
  if len = _0_ then s else
  (* Now, we have len > 0.
     Therefore, in the following, S.unsafe_get will be safe, and Nat.pred
     will not fail. *)
  let s0 = S.unsafe_get s _0_ in
  if s0 <> ' ' && s0 <> '\t' then s else
  chomp (S.sub s _1_ (Nat.pred len))
;;

(* val do_dynamic_link : unit -> unit *)
let do_dynamic_link () =
  let load s =
    if Dynlink.is_native then
      Log.say Log.Warn "Warning: %S was not loaded, running in opt" s else
    try Dynlink.loadfile s with
    | Dynlink.Error e ->
      Log.say Log.Warn
        "Warning: dynamic link error, %s" (Dynlink.error_message e) in
  let rec loop s =
    match
      try Some (S.index s ';') with
      | _ -> None
    with
    | None ->
      let s = chomp s in
      if s <> "" then load s
    | Some i ->
      let fn = S.sub s _0_ i in
      load (chomp fn);
      let i = Nat.succ i in
      let len = S.length s in
      assert (i <= len);
      (* Now, we have i <= len.
         Therefore, the next Nat.sub will not fail. *)
      loop (S.sub s i (Nat.sub len i)) in
  Dynlink.allow_unsafe_modules true;
  loop (try Sys.getenv "SKLML_DLINK" with Not_found -> "")
;;

let start node =
  let debug = debug "start" in
  debug "entering";
  let sl_packet = { color = node.sn_color; } in
  if node.sn_dynamic_link then do_dynamic_link ();
  try
    let (_, oc) as io = IO.connect node.sn_master_address in
    IO.to_channel oc sl_packet true;
    slave_loop io node;
    debug "leaving"
  with
  | IO.IOError msg -> Log.say Log.Err "IO error, %s" msg; exit 1
;;
