(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2007-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux                                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lrope.ml,v 1.22 2017/01/04 11:58:30 fclement Exp $ *)

(* Linear ropes, optimized for big buffers handling *)

module Nat = Unatural;;
module L = Ulist;;
module B = Ubytes;;

let _0_ = Nat._0_;;

let nat_sys_max_string_length = Nat.unsafe_of_int Sys.max_string_length;;

let b_index_from_to s f t c =
  let len = B.length s in
  if len = _0_ then None else
  if t > len then None else
  let rec loop i =
    (* i is increasing from its initial value. *)
    if i >= t then None else
    (* Now, we have 0 <= i < t <= len.
       Therefore, the next B.unsafe_get will be safe. *)
    if B.unsafe_get s i = c then Some i else
    loop (Nat.succ i) in
  loop f
;;

open Tools_common_types;;

type total_length = length;;

(* Invariant for S (s, ofs, len, tot_len, lr):
   0 <= ofs &&
   ofs + len <= length s &&
   len <= tot_len. *)
type t =
  | Nil
  | S of B.t * offset * length * total_length * t Lazy.t
and lrope = t
;;

let length = function
  | Nil -> _0_
  | S (_, _, _, tot_len, _) -> tot_len
;;

let empty = Nil;;

let bytes_append lr s =
  let len = B.length s in
  if len = _0_ then lr else
  let ofs = _0_
  and tot_len = Nat.add (length lr) len in
  (* We have
     0 <= ofs &&
     ofs + len <= length s &&
     len <= tot_len. *)
  S (s, ofs, len, tot_len, lazy lr)
;;

let rec rev_list_append lr = function
  | [] -> lr
  | s :: sl -> rev_list_append (bytes_append lr s) sl
;;

let list_append lr l = rev_list_append lr (L.rev l);;

let index lr c =
  let rec loop d = function
    | Nil -> None
    | S (s, ofs, len, _, lazy lr) ->
      begin match b_index_from_to s ofs (Nat.add ofs len) c with
      | None -> loop (Nat.add d len) lr
      | Some i -> Some (Nat.add d i)
      end in
  loop _0_ lr
;;

let take_n lr n =
  if n >= nat_sys_max_string_length || n > length lr then
    invalid_arg "take_n" else
  let rec loop out off rem = function
    | Nil -> ()
    | S (s, ofs, len, _, lazy lr) ->
      let l = min rem len in
      B.blit s ofs out off l;
      if rem <= l then () else
      (* Now, we have l < rem.
         Therefore, the following Nat.sub will not fail. *)
      loop out (Nat.add off l) (Nat.sub rem l) lr in
  let s = B.create n in
  loop s _0_ n lr;
  s
;;

let skip_n lr n =
  if n >= nat_sys_max_string_length || n > length lr then
    invalid_arg "skip_n" else
  let rec loop rem = function
    | Nil -> Nil
    | S (s, ofs, len, tot_len, lazy lr) ->
      if len <= rem then
        (* Now, we have len <= rem.
           Therefore, the following Nat.sub will not fail. *)
        loop (Nat.sub rem len) lr else
        (* Now, we have rem < len <= l.
           Therefore, both following Nat.sub will not fail. *)
      let ofs = Nat.add ofs rem
      and len = Nat.sub len rem
      and tot_len = Nat.sub tot_len rem in
      (* We have
         0 <= ofs &&
         ofs + len <= length s &&
         len <= tot_len. *)
      S (s, ofs, len, tot_len, lazy lr) in
  loop n lr
;;

let get_c = function
  | Nil -> invalid_arg "get_c"
  | S (s, ofs, _, _, _) ->
    (* From the invariant, we have 0 <= ofs < ofs + len <= length s.
       Therefore, the next B.unsafe_get will be safe. *)
    B.unsafe_get s ofs
;;

let rec concat lr0 lr1 =
  match lr0 with
  | Nil -> lr1
  | S (s, ofs, len, l, lazy lr) ->
    S (s, ofs, len, Nat.add l (length lr1), lazy (concat lr lr1))
;;
